# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""WINDOW tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import ExprLiteral, Query, Table, Window

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.fixture(autouse=True)
async def _autotable(txn: Transaction) -> None:
    tbl = Table('tbl')
    await txn.exq(
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .primary_key(autoincrement=True)
        .unique()
        .column('char', 'TEXT')
        .unique()
        .column('name', 'TEXT'),
    )

    await txn.exq(
        Query()
        .insert(tbl)
        .columns(tbl.char, tbl.name)
        .values(
            ('A', 'one'),
            ('B', 'two'),
            ('C', 'three'),
            ('D', 'one'),
            ('E', 'two'),
            ('F', 'three'),
            ('G', 'one'),
        ),
    )


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',  # does not use txn
    ],
    indirect=True,
)
async def test_syntax() -> None:
    """Test syntax."""
    tbl = Table('tbl')
    assert (Window().execute('lead', tbl.id).over_name('window_name').__getsql__())[0] == (
        'lead("tbl"."id") OVER window_name'
    )

    assert (
        Window()
        .execute('lead', tbl.id)
        .filter(tbl.name == 'one')
        .over_name('window_name')
        .__getsql__()
    )[0] == ('lead("tbl"."id") FILTER (WHERE "tbl"."name" = ?) OVER window_name')

    assert (Window().execute('lead', tbl.id).over_definition().__getsql__())[0] == (
        'lead("tbl"."id") OVER ()'
    )

    assert (
        Window().execute('lead', tbl.id).over_definition().base_window('window_name').__getsql__()
    )[0] == ('lead("tbl"."id") OVER (window_name)')

    assert (
        Window().execute('lead', tbl.id).over_definition().partition_by(tbl.rel_id).__getsql__()
    )[0] == ('lead("tbl"."id") OVER ( PARTITION BY "tbl"."rel_id")')

    assert (Window().execute('lead', tbl.id).over_definition().order_by(tbl.a, tbl.b).__getsql__())[
        0
    ] == ('lead("tbl"."id") OVER ( ORDER BY "tbl"."a", "tbl"."b")')

    assert (Window().execute('lead', tbl.id).over_definition().range().precedeing().__getsql__())[
        0
    ] == ('lead("tbl"."id") OVER ( RANGE UNBOUNDED PRECEDING)')

    assert (
        Window()
        .execute('lead', tbl.id)
        .over_definition()
        .rows()
        .precedeing(ExprLiteral(1))
        .__getsql__()
    )[0] == ('lead("tbl"."id") OVER ( ROWS ? PRECEDING)')

    assert (Window().execute('lead', tbl.id).over_definition().groups().current_row().__getsql__())[
        0
    ] == ('lead("tbl"."id") OVER ( GROUPS CURRENT ROW)')

    assert (
        Window()
        .execute('lead', tbl.id)
        .over_definition()
        .groups()
        .between()
        .precedeing()
        .following(ExprLiteral(1))
        .__getsql__()
    )[0] == ('lead("tbl"."id") OVER ( GROUPS BETWEEN UNBOUNDED PRECEDING AND ? FOLLOWING)')

    assert (
        Window()
        .execute('lead', tbl.id)
        .over_definition()
        .groups()
        .between()
        .precedeing(ExprLiteral(1))
        .following()
        .__getsql__()
    )[0] == ('lead("tbl"."id") OVER ( GROUPS BETWEEN ? PRECEDING AND UNBOUNDED FOLLOWING)')

    assert (
        Window()
        .execute('lead', tbl.id)
        .over_definition()
        .groups()
        .between()
        .current_row()
        .precedeing(ExprLiteral(1))
        .__getsql__()
    )[0] == ('lead("tbl"."id") OVER ( GROUPS BETWEEN CURRENT ROW AND ? PRECEDING)')

    assert (
        Window()
        .execute('lead', tbl.id)
        .over_definition()
        .groups()
        .between()
        .following(ExprLiteral(1))
        .current_row()
        .__getsql__()
    )[0] == ('lead("tbl"."id") OVER ( GROUPS BETWEEN ? FOLLOWING AND CURRENT ROW)')

    assert (
        Window()
        .execute('lead', tbl.id)
        .over_definition()
        .groups()
        .precedeing()
        .exclude_no_others()
        .__getsql__()
    )[0] == ('lead("tbl"."id") OVER ( GROUPS UNBOUNDED PRECEDING EXCLUDE NO OTHERS)')

    assert (
        Window()
        .execute('lead', tbl.id)
        .over_definition()
        .groups()
        .precedeing()
        .exclude_current_row()
        .__getsql__()
    )[0] == ('lead("tbl"."id") OVER ( GROUPS UNBOUNDED PRECEDING EXCLUDE CURRENT ROW)')

    assert (
        Window()
        .execute('lead', tbl.id)
        .over_definition()
        .groups()
        .precedeing()
        .exclude_group()
        .__getsql__()
    )[0] == ('lead("tbl"."id") OVER ( GROUPS UNBOUNDED PRECEDING EXCLUDE GROUP)')

    assert (
        Window()
        .execute('lead', tbl.id)
        .over_definition()
        .groups()
        .precedeing()
        .exclude_ties()
        .__getsql__()
    )[0] == ('lead("tbl"."id") OVER ( GROUPS UNBOUNDED PRECEDING EXCLUDE TIES)')


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='group_concat unsupported')),
    ],
    indirect=True,
)
async def test_filter(txn: Transaction) -> None:
    """Test filter."""
    tbl = Table('tbl')
    query = (
        Query()
        .select(
            tbl.name.typed(str),
            tbl.id.typed(int),
            tbl.char.typed(str),
            Window()
            .execute('group_concat', tbl.char, ExprLiteral('.'))
            .filter(tbl.name != 'two')
            .over_definition()
            .order_by(tbl.id)
            .typed(str),
        )
        .from_(tbl)
    )
    assert query.__getsql__() == (
        'SELECT "tbl"."name", "tbl"."id", "tbl"."char", '
        'group_concat("tbl"."char", ?) FILTER (WHERE "tbl"."name" != ?) '
        'OVER ( ORDER BY "tbl"."id") FROM "tbl"',
        ('.', 'two'),
    )
    assert await txn.exq(query) == [
        ('one', 1, 'A', 'A'),
        ('two', 2, 'B', 'A'),
        ('three', 3, 'C', 'A.C'),
        ('one', 4, 'D', 'A.C.D'),
        ('two', 5, 'E', 'A.C.D'),
        ('three', 6, 'F', 'A.C.D.F'),
        ('one', 7, 'G', 'A.C.D.F.G'),
    ]


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='group_concat unsupported')),
    ],
    indirect=True,
)
async def test_aggregate(txn: Transaction) -> None:
    """Test aggregate."""
    tbl = Table('tbl')
    query = (
        Query()
        .select(
            tbl.id.typed(int),
            tbl.char.typed(str),
            Window()
            .execute('group_concat', tbl.char, ExprLiteral('.'))
            .over_definition()
            .order_by(tbl.id)
            .rows()
            .between()
            .precedeing(ExprLiteral(1))
            .following(ExprLiteral(1))
            .typed(str),
        )
        .from_(tbl)
    )
    assert query.__getsql__() == (
        'SELECT "tbl"."id", "tbl"."char", '
        'group_concat("tbl"."char", ?) OVER ( ORDER BY "tbl"."id" ROWS BETWEEN '
        '? PRECEDING AND ? FOLLOWING) FROM "tbl"',
        ('.', 1, 1),
    )
    assert await txn.exq(query) == [
        (1, 'A', 'A.B'),
        (2, 'B', 'A.B.C'),
        (3, 'C', 'B.C.D'),
        (4, 'D', 'C.D.E'),
        (5, 'E', 'D.E.F'),
        (6, 'F', 'E.F.G'),
        (7, 'G', 'F.G'),
    ]


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='group_concat unsupported')),
    ],
    indirect=True,
)
async def test_partition(txn: Transaction) -> None:
    """Test partition by."""
    tbl = Table('tbl')
    query = (
        Query()
        .select(
            tbl.name.typed(str),
            tbl.id.typed(int),
            tbl.char.typed(str),
            Window()
            .execute('group_concat', tbl.char, ExprLiteral('.'))
            .over_definition()
            .partition_by(tbl.name)
            .order_by(tbl.id)
            .rows()
            .between()
            .current_row()
            .following()
            .typed(str),
        )
        .from_(tbl)
    )
    assert query.__getsql__() == (
        'SELECT "tbl"."name", "tbl"."id", "tbl"."char", '
        'group_concat("tbl"."char", ?) OVER ( PARTITION BY "tbl"."name" ORDER BY "tbl"."id" ROWS '
        'BETWEEN CURRENT ROW AND UNBOUNDED FOLLOWING) FROM "tbl"',
        ('.',),
    )
    assert await txn.exq(query) == [
        ('one', 1, 'A', 'A.D.G'),
        ('one', 4, 'D', 'D.G'),
        ('one', 7, 'G', 'G'),
        ('three', 3, 'C', 'C.F'),
        ('three', 6, 'F', 'F'),
        ('two', 2, 'B', 'B.E'),
        ('two', 5, 'E', 'E'),
    ]

    query_ordered = (
        Query()
        .select(
            tbl.name.typed(str),
            tbl.id.typed(int),
            tbl.char.typed(str),
            Window()
            .execute('group_concat', tbl.char, ExprLiteral('.'))
            .over_definition()
            .partition_by(tbl.name)
            .order_by(tbl.id)
            .rows()
            .between()
            .current_row()
            .following()
            .typed(str),
        )
        .from_(tbl)
        .order_by(tbl.id)
    )
    assert query_ordered.__getsql__() == (
        'SELECT "tbl"."name", "tbl"."id", "tbl"."char", '
        'group_concat("tbl"."char", ?) OVER ( PARTITION BY "tbl"."name" ORDER BY "tbl"."id" ROWS '
        'BETWEEN CURRENT ROW AND UNBOUNDED FOLLOWING) FROM "tbl" ORDER BY "tbl"."id"',
        ('.',),
    )
    assert await txn.exq(query_ordered) == [
        ('one', 1, 'A', 'A.D.G'),
        ('two', 2, 'B', 'B.E'),
        ('three', 3, 'C', 'C.F'),
        ('one', 4, 'D', 'D.G'),
        ('two', 5, 'E', 'E'),
        ('three', 6, 'F', 'F'),
        ('one', 7, 'G', 'G'),
    ]



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL window expressions."""

from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from sqls.queries.base import QBase, Term
from sqls.queries.expr import Expr, ExprTyped

if TYPE_CHECKING:
    from sqls.interfaces import BindArg

T = TypeVar('T')


class Window(QBase[tuple[()]]):
    """WINDOW function mixin."""

    def execute(self, funcname: str, *exprs: Expr) -> PostExecute:
        """Start WINDOW function invocation.

        Args:
            funcname: Function name.
            exprs: Function arguments.

        Returns:
            Query builder.

        """
        sqls = []
        args: list[BindArg] = []
        for expr in exprs:
            sql, argz = expr.__getsql__()
            sqls.append(sql)
            args += argz
        return self._f(PostExecute, f'{funcname}({", ".join(sqls)})', tuple(args))


class Filter(QBase[tuple[()]]):
    """FILTER mixin."""

    def filter(self, expr: Expr) -> Over:
        """Add FILTER clause.

        Args:
            expr: Expression.

        Returns:
            Query builder.

        """
        sql, args = expr.__getsql__()
        return self._f(Over, f' FILTER (WHERE {sql})', args)


class Over(QBase[tuple[()]]):
    """OVER mixin."""

    def over_name(self, name: str) -> WindowTerm:
        """Add OVER window name clause.

        Args:
            name: Window name.

        Returns:
            Query builder.

        """
        return self._f(WindowTerm, f' OVER {name}', ())

    def over_definition(self) -> PostOver:
        """Add OVER window definition clause.

        Returns:
            Query builder.

        """
        enter = ((Term,), ')')
        return self._f(PostOver, ' OVER (', enter=enter)


class BaseWindow(QBase[tuple[()]]):
    """Base window mixin."""

    def base_window(self, name: str) -> PostBaseWindow:
        """Add base window name.

        Copies partition and order from base window.

        Args:
            name: Window name.

        Returns:
            Query builder.

        """
        return self._f(PostBaseWindow, f'{name}', ())


class WindowTerm(QBase[tuple[()]], Expr):
    """Window term mixin."""

    def endq(self) -> Term[tuple[()]]:
        """Mark window function statement done."""
        return Term(self)

    def typed(self, _: type[T]) -> ExprTyped[T]:
        """Convert function invocation to typed expression.

        Args:
            _: Type argument.

        Returns:
            Typed column expression.

        """
        return ExprTyped[T](self)


class PartitionBy(QBase[tuple[()]]):
    """PARTITION BY mixin."""

    def partition_by(self, *exprs: Expr) -> PostPartitionBy:
        """Add PARTITION BY clause.

        Args:
            exprs: Expressions.

        Returns:
            Query builder.

        """
        sqls = []
        args: list[BindArg] = []
        for expr in exprs:
            sql, argz = expr.__getsql__()
            sqls.append(sql)
            args += argz
        return self._f(PostPartitionBy, f' PARTITION BY {", ".join(sqls)}', tuple(args))


class OrderBy(QBase[tuple[()]]):
    """ORDER BY mixin."""

    def order_by(
        self,
        *exprs: Expr,
        collate: str | None = None,
        desc: bool = False,
        nulls: str | None = None,
    ) -> PostOrderBy:
        """Add ORDER BY clause.

        Args:
            exprs: Expressions.
            collate: Collation.
            desc: Sort descending.
            nulls: NULLS keyword to add.

        Returns:
            Query builder.

        """
        sqls = []
        args: list[BindArg] = []
        for expr in exprs:
            sql, argz = expr.__getsql__()
            sqls.append(sql)
            args += argz
        collation = f' COLLATE {collate}' if collate else ''
        sort = ' DESC' if desc else ''
        nulls = '' if nulls is None else f' NULLS {nulls}'
        return self._f(
            PostOrderBy,
            f' ORDER BY {", ".join(sqls)}{collation}{sort}{nulls}',
            tuple(args),
        )


class FrameType(QBase[tuple[()]]):
    """Frame type mixin."""

    def range(self) -> FrameBoundary:
        """Add RANGE clause."""
        return self._f(FrameBoundary, ' RANGE', ())

    def rows(self) -> FrameBoundary:
        """Add ROWS clause."""
        return self._f(FrameBoundary, ' ROWS', ())

    def groups(self) -> FrameBoundary:
        """Add GROUPS clause."""
        return self._f(FrameBoundary, ' GROUPS', ())


class FrameBoundary(QBase[tuple[()]]):
    """Frame boundary."""

    def between(self) -> FrameBoundaryStart:
        """Start BETWEEN clause."""
        return self._f(FrameBoundaryStart, ' BETWEEN', ())

    def precedeing(self, expr: Expr | None = None) -> PostFrameBoundary:
        """Add PRECEDING clause.

        Args:
            expr: Expression.

        Returns:
            Query builder.

        """
        sql = ''
        args: list[BindArg] = []
        if expr is None:
            sql += ' UNBOUNDED PRECEDING'
        else:
            subsql, subargs = expr.__getsql__()
            sql += f' {subsql} PRECEDING'
            args += subargs
        return self._f(PostFrameBoundary, sql, tuple(args))

    def current_row(self) -> PostFrameBoundary:
        """Add CURRENT ROW clause."""
        return self._f(PostFrameBoundary, ' CURRENT ROW', ())


class FrameBoundaryStart(QBase[tuple[()]]):
    """Frame boundary start mixin."""

    def precedeing(self, expr: Expr | None = None) -> FrameBoundaryEnd:
        """Add PRECEDING clause for frame start.

        Args:
            expr: Expression or None for UNBOUNDED.

        Returns:
            Query builder.

        """
        sql = ''
        args: list[BindArg] = []
        if expr is None:
            sql += ' UNBOUNDED PRECEDING AND'
        else:
            subsql, subargs = expr.__getsql__()
            sql += f' {subsql} PRECEDING AND'
            args += subargs
        return self._f(FrameBoundaryEnd, sql, tuple(args))

    def current_row(self) -> FrameBoundaryEnd:
        """Add CURRENT ROW clause for frame start."""
        return self._f(FrameBoundaryEnd, ' CURRENT ROW AND', ())

    def following(self, expr: Expr) -> FrameBoundaryEnd:
        """Add FOLLOWING clause for frame start.

        Args:
            expr: Expression.

        Returns:
            Query builder.

        """
        sql, args = expr.__getsql__()
        return self._f(FrameBoundaryEnd, f' {sql} FOLLOWING AND', args)


class FrameBoundaryEnd(QBase[tuple[()]]):
    """Frame boundary end mixin."""

    def precedeing(self, expr: Expr) -> PostFrameBoundary:
        """Add PRECEDING clause for frame end.

        Args:
            expr: Expression.

        Returns:
            Query builder.

        """
        sql, args = expr.__getsql__()
        return self._f(PostFrameBoundary, f' {sql} PRECEDING', args)

    def current_row(self) -> PostFrameBoundary:
        """Add CURRENT ROW clause for frame end."""
        return self._f(PostFrameBoundary, ' CURRENT ROW', ())

    def following(self, expr: Expr | None = None) -> PostFrameBoundary:
        """Add FOLLOWING clause for frame end.

        Args:
            expr: Expression or None for UNBOUNDED.

        Returns:
            Query builder.

        """
        sql = ''
        args: list[BindArg] = []
        if expr is None:
            sql += ' UNBOUNDED FOLLOWING'
        else:
            subsql, subargs = expr.__getsql__()
            sql += f' {subsql} FOLLOWING'
            args += subargs
        return self._f(PostFrameBoundary, sql, tuple(args))


class Exclude(QBase[tuple[()]]):
    """EXCLUDE mixin."""

    def exclude_no_others(self) -> WindowTerm:
        """Add EXCLUDE NO OTHERS clause."""
        return self._f(WindowTerm, ' EXCLUDE NO OTHERS', ())

    def exclude_current_row(self) -> WindowTerm:
        """Add EXCLUDE CURRENT ROW clause."""
        return self._f(WindowTerm, ' EXCLUDE CURRENT ROW', ())

    def exclude_group(self) -> WindowTerm:
        """Add EXCLUDE GROUP clause."""
        return self._f(WindowTerm, ' EXCLUDE GROUP', ())

    def exclude_ties(self) -> WindowTerm:
        """Add EXCLUDE TIES clause."""
        return self._f(WindowTerm, ' EXCLUDE TIES', ())


class PostFrameBoundary(
    Exclude,
    WindowTerm,
):
    """Post frame boundary builder."""


class PostExecute(
    Filter,
    Over,
):
    """Post execute builder."""


class PostOver(
    BaseWindow,
    PartitionBy,
    OrderBy,
    FrameType,
    WindowTerm,
):
    """Post over builder."""


class PostBaseWindow(
    PartitionBy,
    OrderBy,
    FrameType,
    WindowTerm,
):
    """Post basewindow builder."""


class PostPartitionBy(
    OrderBy,
    FrameType,
    WindowTerm,
):
    """Post partitionby builder."""


class PostOrderBy(
    FrameType,
    WindowTerm,
):
    """Post orderby builder."""



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Declarative models and code generators."""

from __future__ import annotations

from .fields import (
    AutoField,
    BigAutoField,
    BigIntegerField,
    BlobField,
    BooleanField,
    CharField,
    DateField,
    DateTimeField,
    DecimalField,
    DoubleField,
    FixedCharField,
    FloatField,
    ForeignKeyField,
    IntegerField,
    SmallIntegerField,
    TextField,
    TimeField,
    UUIDField,
)
from .generators import get_create_queries
from .model import Fieldmeta, Indexmeta, Model, Tablemeta

__all__ = [
    'AutoField',
    'BigAutoField',
    'BigIntegerField',
    'BlobField',
    'BooleanField',
    'CharField',
    'DateField',
    'DateTimeField',
    'DecimalField',
    'DoubleField',
    'Fieldmeta',
    'FixedCharField',
    'FloatField',
    'ForeignKeyField',
    'Indexmeta',
    'IntegerField',
    'Model',
    'SmallIntegerField',
    'Tablemeta',
    'TextField',
    'TimeField',
    'UUIDField',
    'get_create_queries',
]



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Transaction exceptions."""


class IntegrityError(Exception):
    """IntegrityError from database."""


class OperationalError(Exception):
    """OperationalError from database."""



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""PostgreSQL manager."""

from __future__ import annotations

from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, cast
from urllib.parse import urlparse

import asyncpg  # type: ignore[import-untyped]

from .exceptions import IntegrityError, OperationalError

if TYPE_CHECKING:
    import sys
    from collections.abc import AsyncGenerator
    from types import TracebackType
    from typing import TypeVar

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    from sqls.interfaces import BindArg, ManagerOptions, Query, SqlValue, SupportsGetsql

    T = TypeVar('T')


class Transaction:
    """Transaction class."""

    def __init__(self, connection: asyncpg.Connection) -> None:
        """Initialzize transaction object."""
        self.connection = connection

    async def execute(self, query: str, *args: BindArg) -> list[tuple[SqlValue, ...]]:
        """Execute query."""
        return await self.connection.fetch(query, *args)  # type: ignore[no-any-return]

    async def exq(self, query: Query[T]) -> list[T]:
        """Execute shorthand."""
        qstr, args = query.__getsql__()

        idx = 1
        while '?' in qstr:
            if isinstance(args[idx - 1], int):
                subtyp = '::int'
            elif isinstance(args[idx - 1], float):
                subtyp = '::float'
            else:
                subtyp = ''

            qstr = qstr.replace('?', f'${idx}{subtyp}', 1)
            idx += 1

        # fixups
        qstr = qstr.replace('"id" INTEGER', '"id" SERIAL')
        qstr = qstr.replace(' AUTOINCREMENT', '')
        qstr = qstr.replace(' INTEGER DEFAULT TRUE', 'BOOLEAN DEFAULT TRUE')
        qstr = qstr.replace(' INTEGER DEFAULT FALSE', 'BOOLEAN DEFAULT FALSE')
        qstr = qstr.replace(' VIRTUAL', ' STORED')
        qstr = qstr.replace('REINDEX', 'REINDEX TABLE')
        qstr = qstr.replace('NOT REGEXP', '!~')
        qstr = qstr.replace('REGEXP', '~')
        qstr = qstr.replace('group_concat', 'string_agg')

        try:
            return await self.connection.fetch(qstr, *args)  # type: ignore[no-any-return]
        except asyncpg.exceptions.UniqueViolationError as err:
            msg = f'QUERY: {query.__getsql__()!r}, ERROR {err.args!r}.'
            raise IntegrityError(msg) from None
        except asyncpg.exceptions.PostgresError as err:
            msg = f'QUERY: {query.__getsql__()!r}, ERROR {err.args!r}.'
            raise OperationalError(msg) from None

    async def exq_count(self, query: SupportsGetsql) -> int:
        """Execute shorthand."""
        return len(await self.exq(cast('Query[None]', query)))

    async def __aenter__(self) -> Self:
        """Transaction setup."""
        await self.connection.execute(f'SAVEPOINT "{id(self)}"')
        return self.__class__(self.connection)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        """Transaction cleanup."""
        if exc is not None:
            await self.connection.execute(f'ROLLBACK TO SAVEPOINT "{id(self)}"')
            raise exc
        await self.connection.execute(f'RELEASE SAVEPOINT "{id(self)}"')
        return False


class Manager:
    """Database manager."""

    def __init__(self, uri: str, options: ManagerOptions | None = None) -> None:
        """Initialize."""
        self.uri = uri
        if not options:
            options = cast('ManagerOptions', {})
        self.max_connections = options.get('max_connections', 4)
        self.pool = None

    async def init(self) -> None:
        """Initialize pool."""
        parsed = urlparse(self.uri)
        self.pool = await asyncpg.create_pool(
            host=parsed.hostname,
            port=parsed.port,
            user=parsed.username,
            password=parsed.password,
            database=parsed.path[1:],
            min_size=min(self.max_connections, 10),
            max_size=self.max_connections,
        )

    async def close(self) -> None:
        """Close all database connections."""
        assert self.pool
        await self.pool.close()

    @asynccontextmanager
    async def txn(self, _: bool = False) -> AsyncGenerator[Transaction, None]:
        """Open a transaction for database operations."""
        assert self.pool
        async with self.pool.acquire() as connection, connection.transaction():
            yield Transaction(connection)



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL expressions."""

from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

from .exceptions import QueryError

if TYPE_CHECKING:
    from typing import TypeAlias

    from sqls.interfaces import BindArg, Query, SqlValue

    Other: TypeAlias = 'tuple[SqlValue, ...] | SqlValue | Expr'

T = TypeVar('T')


def fmt(value: SqlValue) -> str:
    """Format value for SQL string.

    Args:
        value: Value to format.

    Returns:
        Formatted value.

    Raises:
        QueryError: If bytes passed as value.

    """
    if isinstance(value, int | float):
        return repr(value)

    if isinstance(value, bytes):  # pragma: no cover
        raise QueryError

    assert isinstance(value, str)
    value = value.replace("'", "''")
    return f"'{value}'"


class ExprTyped(Generic[T]):
    """Typed Expression."""

    def __init__(self, child: Expr | ExprAlias) -> None:
        """Initialize.

        Args:
            child: Expression.

        """
        super().__init__()
        self.child = child

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        return self.child.__getsql__()


class Expr:
    """SQL Expression."""

    def positive(self) -> ExprUnary:
        """Make positive."""
        return ExprUnary('+', self)

    def negative(self) -> ExprUnary:
        """Make negative."""
        return ExprUnary('-', self)

    def bitwise_negate(self) -> ExprUnary:
        """Bitwise negation."""
        return ExprUnary('~', self)

    def negate(self) -> ExprUnary:
        """Negate expression."""
        return ExprUnary('NOT', self)

    def strconcat(self, other: Other) -> ExprBinary:
        """Concatenate other string."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('||', self, other)

    def __mul__(self, other: Other) -> ExprBinary:
        """Multiply."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('*', self, other)

    def __truediv__(self, other: Other) -> ExprBinary:
        """Divide."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('/', self, other)

    def __mod__(self, other: Other) -> ExprBinary:
        """Modulo."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('%', self, other)

    def __add__(self, other: Other) -> ExprBinary:
        """Add."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('+', self, other)

    def __sub__(self, other: Other) -> ExprBinary:
        """Subtract."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('-', self, other)

    def __lshift__(self, other: Other) -> ExprBinary:
        """Left shift."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('<<', self, other)

    def __rshift__(self, other: Other) -> ExprBinary:
        """Right shift."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('>>', self, other)

    def __and__(self, other: Other) -> ExprBinary:
        """Binary and."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('&', self, other)

    def __or__(self, other: Other) -> ExprBinary:
        """Binary or."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('|', self, other)

    def __lt__(self, other: Other) -> ExprBinary:
        """Compare lt."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('<', self, other)

    def __le__(self, other: Other) -> ExprBinary:
        """Compare le."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('<=', self, other)

    def __gt__(self, other: Other) -> ExprBinary:
        """Compare gt."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('>', self, other)

    def __ge__(self, other: Other) -> ExprBinary:
        """Compare ge."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('>=', self, other)

    def __eq__(self, other: Other) -> ExprBinary:  # type: ignore[override]
        """Compare eq."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('=', self, other)

    def __ne__(self, other: Other) -> ExprBinary:  # type: ignore[override]
        """Compare ne."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('!=', self, other)

    def and_(self, other: Other) -> ExprBinary:
        """Compare and."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('AND', self, other)

    def or_(self, other: Other) -> ExprBinary:
        """Compare or."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('OR', self, other)

    def cast(self, typ: str) -> ExprFlexible:
        """Cast."""
        return ExprFlexible('CAST (', self, f' AS {typ})')

    def _match(self, match: str, expr: Other, escape: str | None = None) -> ExprFlexible:
        """Create match expression and optional escape.

        Args:
            match: Match function.
            expr: Expression to match.
            escape: Escape character.

        Returns:
            Match Expression.

        """
        if not isinstance(expr, Expr):
            expr = ExprLiteral(expr)
        esc = (f' ESCAPE {escape!r}',) if escape else ()
        return ExprFlexible(self, f' {match} ', expr, *esc)

    def like(self, expr: Other, escape: str | None = None) -> ExprFlexible:
        """Like."""
        return self._match('LIKE', expr, escape)

    def not_like(self, expr: Other, escape: str | None = None) -> ExprFlexible:
        """Not like."""
        return self._match('NOT LIKE', expr, escape)

    def glob(self, expr: Other, escape: str | None = None) -> ExprFlexible:
        """Glob."""
        return self._match('GLOB', expr, escape)

    def not_glob(self, expr: Other, escape: str | None = None) -> ExprFlexible:
        """Not glob."""
        return self._match('NOT GLOB', expr, escape)

    def regexp(self, expr: Other, escape: str | None = None) -> ExprFlexible:
        """Regexp."""
        return self._match('REGEXP', expr, escape)  # pragma: no cover

    def not_regexp(self, expr: Other, escape: str | None = None) -> ExprFlexible:
        """Not regexp."""
        return self._match('NOT REGEXP', expr, escape)  # pragma: no cover

    def match(self, expr: Other, escape: str | None = None) -> ExprFlexible:
        """Match."""
        return self._match('MATCH', expr, escape)  # pragma: no cover

    def not_match(self, expr: Other, escape: str | None = None) -> ExprFlexible:
        """Not match."""
        return self._match('NOT MATCH', expr, escape)  # pragma: no cover

    def is_null(self) -> ExprFlexible:
        """Is null."""
        return ExprFlexible(self, ' IS NULL')

    def not_null(self) -> ExprFlexible:
        """Is not null."""
        return ExprFlexible(self, ' IS NOT NULL')

    def is_(self, other: Other) -> ExprBinary:
        """Compare is."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('IS', self, other)

    def is_not(self, other: Other) -> ExprBinary:
        """Compare is not."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('IS NOT', self, other)

    def between(self, expr1: Other, expr2: Other) -> ExprFlexible:
        """Between."""
        if not isinstance(expr1, Expr):
            expr1 = ExprLiteral(expr1)
        if not isinstance(expr2, Expr):
            expr2 = ExprLiteral(expr2)
        return ExprFlexible(self, ' BETWEEN ', expr1, ' AND ', expr2)

    def not_between(self, expr1: Other, expr2: Other) -> ExprFlexible:
        """Not between."""
        if not isinstance(expr1, Expr):
            expr1 = ExprLiteral(expr1)
        if not isinstance(expr2, Expr):
            expr2 = ExprLiteral(expr2)
        return ExprFlexible(self, ' NOT BETWEEN ', expr1, ' AND ', expr2)

    def in_(self, other: Other) -> ExprBinary:
        """Compare in."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('IN', self, other)

    def not_in(self, other: Other) -> ExprBinary:
        """Compare in."""
        if not isinstance(other, Expr):
            other = ExprLiteral(other)
        return ExprBinary('NOT IN', self, other)

    def as_(self, alias: str) -> ExprAlias:
        """Alias expression."""
        return ExprAlias(self, alias)

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        raise NotImplementedError


class ExprAlias:
    """Expression containg aliased expression."""

    def __init__(self, expr: Expr, alias: str) -> None:
        """Initialize."""
        super().__init__()
        self.expr = expr
        self.alias = alias

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        sql, args = self.expr.__getsql__()
        return f'{sql} AS "{self.alias}"', args

    def typed(self, _: type[T]) -> ExprTyped[T]:
        """Convert to typed."""
        return ExprTyped[T](self)


class ExprSubq(Expr, Generic[T]):
    """Expression containg subquery."""

    def __init__(self, value: Query[T]) -> None:
        """Initialize."""
        super().__init__()
        self.value = value

    def exists(self) -> ExprFlexible:
        """Convert to EXISTS clause."""
        return ExprFlexible('EXISTS ', self)

    def not_exists(self) -> ExprFlexible:
        """Convert to NOT EXISTS clause."""
        return ExprFlexible('NOT EXISTS ', self)

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        sql, args = self.value.__getsql__()
        return f'({sql})', args


class ExprLiteral(Expr):
    """Expression literal."""

    escapechr: str = '?'

    def __init__(
        self,
        value: SqlValue | tuple[SqlValue | Expr, ...],
        bind: bool = True,
    ) -> None:
        """Initialize expression literal."""
        super().__init__()
        self.value = value
        self.bind = bind

    def __str__(self) -> str:
        """Get string value."""
        return f'Literal: {self.value!r}'

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        value = self.value
        if value is None:
            return 'NULL', ()
        if value is True:
            return 'TRUE', ()
        if value is False:
            return 'FALSE', ()

        if isinstance(value, tuple):
            sqls = []
            args: list[BindArg] = []
            for val in value:
                if isinstance(val, Expr):
                    sql, argz = val.__getsql__()
                    sqls.append(sql)
                    args += argz
                elif self.bind:
                    sqls.append(self.escapechr)
                    args.append(val)
                else:
                    sqls.append(fmt(val))
            return f'({", ".join(sqls)})', tuple(args)

        if self.bind:
            assert not isinstance(self.value, tuple | Expr)
            return self.escapechr, (self.value,)
        return fmt(value), ()

    def typed(self, _: type[T]) -> ExprTyped[T]:
        """Convert to typed."""
        return ExprTyped[T](self)


class ExprColumn(Expr):
    """Expression table column."""

    def __init__(self, schema: str | None, table: str, col: str) -> None:
        """Initialize table column expression."""
        super().__init__()
        self.schema = schema
        self.table = table
        self.col = col

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        if self.schema:
            schema = self.schema.replace('"', '""')
            schema = f'"{schema}".'
        else:
            schema = ''

        if self.table:
            table = self.table.replace('"', '""')
            table = f'"{table}".'
        else:
            table = ''

        if self.col == '*':
            col = '*'
        else:
            col = self.col.replace('"', '""')
            col = f'"{col}"'
        return f'{schema}{table}{col}', ()

    def typed(self, _: type[T]) -> ExprTyped[T]:
        """Convert to typed."""
        return ExprTyped[T](self)


class ExprUnary(Expr):
    """Unary expression."""

    def __init__(self, operator: str, expr: Expr) -> None:
        """Initialize unary expression."""
        super().__init__()
        self.operator = operator
        self.expr = expr

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        expr, args = self.expr.__getsql__()
        if not isinstance(self.expr, ExprLiteral | ExprColumn):
            expr = f'({expr})'
        return f'{self.operator} {expr}', args

    def typed(self, _: type[T]) -> ExprTyped[T]:
        """Convert to typed."""
        return ExprTyped[T](self)


class ExprBinary(Expr):
    """Binary expression."""

    def __init__(self, operator: str, left: Expr, right: Expr) -> None:
        """Initialize binary expression."""
        super().__init__()
        self.operator = operator
        self.left = left
        self.right = right

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        left, argsleft = self.left.__getsql__()
        right, argsright = self.right.__getsql__()
        if not isinstance(self.left, ExprLiteral | ExprColumn | ExprSubq):
            left = f'({left})'
        if not isinstance(self.right, ExprLiteral | ExprColumn | ExprSubq):
            right = f'({right})'
        return f'{left} {self.operator} {right}', (*argsleft, *argsright)

    def typed(self, _: type[T]) -> ExprTyped[T]:
        """Convert to typed."""
        return ExprTyped[T](self)


class ExprFunction(Expr):
    """Expression function."""

    def __init__(self, funcname: str, *exprs: Expr) -> None:
        """Initialize function expression."""
        super().__init__()
        self.funcname = funcname
        self.exprs = exprs

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        sqls = []
        args: list[BindArg] = []
        for expr in self.exprs:
            sql, argz = expr.__getsql__()
            sqls.append(sql)
            args += argz
        return f'{self.funcname}({", ".join(sqls)})', tuple(args)

    def typed(self, _: type[T]) -> ExprTyped[T]:
        """Convert to typed."""
        return ExprTyped[T](self)


class ExprFlexible(Expr):
    """Flexible expression."""

    def __init__(self, *args: str | Expr) -> None:
        """Initialize."""
        super().__init__()
        self.args = args

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        res = []
        args: list[BindArg] = []
        for arg in self.args:
            if isinstance(arg, Expr):
                sql, argz = arg.__getsql__()
                args += argz
                if not isinstance(arg, ExprLiteral | ExprColumn | ExprSubq):
                    sql = f'({sql})'
                res.append(sql)
            else:
                res.append(arg)
        return ''.join(res), tuple(args)

    def typed(self, _: type[T]) -> ExprTyped[T]:
        """Convert to typed."""
        return ExprTyped[T](self)



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL UPDATE."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .base import QBase, Term
from .exceptions import QueryError
from .expr import Expr, ExprColumn, ExprLiteral

if TYPE_CHECKING:
    from collections.abc import Iterable

    from sqls.interfaces import BindArg

    from .table import Table, Tabular


class Update(QBase[tuple[()]]):
    """UPDATE statement."""

    def _update(self, cmd: str, table: Table) -> PostUpdate:
        """Start UPDATE or 'action' helper.

        Args:
            cmd: Action.
            table: Target table.

        Returns:
            Query builder.

        """
        value, _ = table.__getsql__()
        return self._f(PostUpdate, f'UPDATE {cmd}{value}')

    def update(self, table: Table) -> PostUpdate:
        """Start UPDATE statement."""
        return self._update('', table)

    def update_or_rollback(self, table: Table) -> PostUpdate:
        """Start UPDATE OR ROLLBACK statement."""
        return self._update('OR ROLLBACK ', table)

    def update_or_abort(self, table: Table) -> PostUpdate:
        """Start UPDATE OR ABORT statement."""
        return self._update('OR ABORT ', table)

    def update_or_replace(self, table: Table) -> PostUpdate:
        """Start UPDATE OR REPLACE statement."""
        return self._update('OR REPLACE ', table)

    def update_or_fail(self, table: Table) -> PostUpdate:
        """Start UPDATE OR FAIL statement."""
        return self._update('OR FAIL ', table)

    def update_or_ignore(self, table: Table) -> PostUpdate:
        """Start UPDATE OR IGNORE statement."""
        return self._update('OR IGNORE ', table)


class Set(QBase[tuple[()]]):
    """SET mixin."""

    def set(
        self,
        cols: Iterable[ExprColumn],
        vals: Iterable[Expr | bool | bytes | int | float | str | None],
    ) -> PostSet:
        """Add SET clause.

        Args:
            cols: Colums.
            vals: Values.

        Returns:
            Query builder.

        Raises:
            QueryError: If column and value lengths differ.


        """
        cols = tuple(cols)
        vals = tuple(vals)
        if not cols or len(cols) != len(vals):
            msg = 'Columns and values need to have same length.'
            raise QueryError(msg)

        sqls = []
        args: list[BindArg] = []
        for val in vals:
            valex = val if isinstance(val, Expr) else ExprLiteral(val)
            sql, argz = valex.__getsql__()
            sqls.append(sql)
            args += argz

        assigns = [f'"{k.col}"={v}' for k, v in zip(cols, sqls, strict=True)]

        return self._f(PostSet, f' SET {", ".join(assigns)}', tuple(args))


class From(QBase[tuple[()]]):
    """FROM mixin."""

    def from_(self, table: Tabular) -> PostFrom:
        """Add FROM clause.

        Args:
            table: Table or subquery.

        Returns:
            Query builder.

        """
        sql, args = table.__getsql__()
        return self._f(PostFrom, f' FROM {sql}', args)


class Where(QBase[tuple[()]]):
    """WHERE mixin."""

    def where(self, expr: Expr) -> PostWhere:
        """Add WHERE clause.

        Args:
            expr: Expression.

        Returns:
            Query builder.

        """
        sql, args = expr.__getsql__()
        return self._f(PostWhere, f' WHERE {sql}', args)


class PostUpdate(
    Set,
):
    """Post update builder."""


class PostSet(
    From,
    Where,
    Term[tuple[()]],
):
    """Post set builder."""


class PostFrom(
    Where,
    Term[tuple[()]],
):
    """Post from builder."""


class PostWhere(
    Term[tuple[()]],
):
    """Post where builder."""



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQLS interface definitions."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Protocol, TypeAlias, TypedDict, TypeVar

if TYPE_CHECKING:
    from contextlib import AbstractAsyncContextManager

SqlValue: TypeAlias = bool | bytes | float | int | str | None
BindArg: TypeAlias = SqlValue | Sequence[SqlValue]

T = TypeVar('T')


class ManagerOptions(TypedDict, total=False):
    """Manager options."""

    max_connections: int
    init_pragmas: Sequence[str]
    fini_pragmas: Sequence[str]


class SupportsGetsql(Protocol):
    """Query Protocol."""

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:  # pragma: no cover
        """Get query string."""
        raise NotImplementedError


class Query(SupportsGetsql, Protocol[T]):
    """Query Protocol."""

    attr: T | None


class Transaction(Protocol):
    """Transaction."""

    async def execute(
        self,
        query: str,
        *args: BindArg,
    ) -> list[tuple[SqlValue, ...]]:  # pragma: no cover
        """Just execute."""
        raise NotImplementedError

    async def exq(self, query: Query[T]) -> list[T]:  # pragma: no cover
        """Execute shorthand."""
        raise NotImplementedError

    async def exq_count(self, query: SupportsGetsql) -> int:  # pragma: no cover
        """Execute shorthand."""
        raise NotImplementedError


class Manager(Protocol):
    """Database manager."""

    def __init__(
        self,
        uri: str,
        options: ManagerOptions | None = None,
    ) -> None:  # pragma: no cover
        """Initialize."""
        raise NotImplementedError

    async def init(self) -> None:  # pragma: no cover
        """Initialize database connections."""
        raise NotImplementedError

    async def close(self) -> None:  # pragma: no cover
        """Close all database connections."""
        raise NotImplementedError

    def txn(
        self,
        immediate: bool = False,
    ) -> AbstractAsyncContextManager[Transaction]:  # pragma: no cover
        """Open a transaction for database operations."""
        raise NotImplementedError



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""EXPLAIN tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import Explain, Query, Table

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='no vaccum')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='no vacuum')),
    ],
    indirect=True,
)
async def test_vacuum(txn: Transaction) -> None:
    """Test explain statement."""
    query = Explain().vacuum()
    assert query.__getsql__() == ('EXPLAIN VACUUM', ())
    assert (
        [
            x[1]  # type: ignore[misc]
            for x in await txn.exq(query)
        ][:3]
        == ['Init', 'Vacuum', 'Halt']
    )  # PY3.8: no 'Goto'


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='no query plan')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='no query plan')),
    ],
    indirect=True,
)
async def test_query_plan(txn: Transaction) -> None:
    """Test explain statement."""
    tbl = Table('tbl')
    await txn.exq(Query().create_table(tbl).column('id', 'INTEGER'))

    query = Explain(query_plan=True).select(tbl.id.typed(int)).from_(tbl)
    assert query.__getsql__() == ('EXPLAIN QUERY PLAN SELECT "tbl"."id" FROM "tbl"', ())
    assert await txn.exq(
        query,
    ) in (  # type: ignore[comparison-overlap]
        [(2, 0, 0, 'SCAN tbl')],
        [(2, 0, 0, 'SCAN TABLE tbl')],  # PY3.10
    )


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        'mysql',
    ],
    indirect=True,
)
async def test_query(txn: Transaction) -> None:
    """Test explain statement."""
    tbl = Table('tbl')
    await txn.exq(Query().create_table(tbl).column('id', 'INTEGER'))

    query = Explain().select(tbl.id.typed(int)).from_(tbl)
    assert query.__getsql__() == ('EXPLAIN SELECT "tbl"."id" FROM "tbl"', ())
    assert await txn.exq(query) in (
        [
            (0, 'Init', 0, 7, 0, None, 0, None),
            (1, 'OpenRead', 0, 2, 0, '1', 0, None),
            (2, 'Rewind', 0, 6, 0, None, 0, None),
            (3, 'Column', 0, 0, 1, None, 0, None),
            (4, 'ResultRow', 1, 1, 0, None, 0, None),
            (5, 'Next', 0, 3, 0, None, 1, None),
            (6, 'Halt', 0, 0, 0, None, 0, None),
            (7, 'Transaction', 0, 0, 1, '0', 1, None),
            (8, 'Goto', 0, 1, 0, None, 0, None),
        ],
        [('Seq Scan on tbl  (cost=0.00..35.50 rows=2550 width=4)',)],
        [(1, 'SIMPLE', 'tbl', 'ALL', None, None, None, None, '1', '')],
    )



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""MySQL manager."""

from __future__ import annotations

from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, cast
from urllib.parse import urlparse

import aiomysql  # type: ignore[import-untyped]

from .exceptions import IntegrityError, OperationalError

if TYPE_CHECKING:
    import sys
    from collections.abc import AsyncIterator
    from types import TracebackType
    from typing import TypeVar

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    from sqls.interfaces import BindArg, ManagerOptions, Query, SqlValue, SupportsGetsql

    T = TypeVar('T')


class Transaction:
    """Transaction class."""

    def __init__(self, cursor: aiomysql.Cursor) -> None:
        """Initialzize transaction object."""
        self.cursor = cursor

    async def execute(self, query: str, *args: BindArg) -> list[tuple[SqlValue, ...]]:
        """Execute query."""
        return await self.cursor.execute(query, args)  # type: ignore[no-any-return]

    async def exq(self, query: Query[T]) -> list[T]:
        """Execute shorthand."""
        qstr, args = query.__getsql__()
        qstr = qstr.replace('?', '%s')

        # fixups
        qstr = qstr.replace('AUTOINCREMENT', 'AUTO_INCREMENT')
        qstr = qstr.replace('DEFAULT VALUES', 'VALUES ()')
        qstr = qstr.replace('INSERT OR IGNORE', 'INSERT IGNORE')
        qstr = qstr.replace('UPDATE OR IGNORE', 'UPDATE IGNORE')
        qstr = qstr.replace('AS TEXT', 'AS CHAR')
        qstr = qstr.replace(' % ', ' MOD ')

        try:
            await self.cursor.execute(qstr, args)
        except aiomysql.IntegrityError as err:
            msg = f'QUERY: {query.__getsql__()!r}, ERROR {err.args!r}.'
            raise IntegrityError(msg) from None
        except aiomysql.Error as err:
            msg = f'QUERY: {query.__getsql__()!r}, ERROR {err.args!r}.'
            raise OperationalError(msg) from None

        return list(await self.cursor.fetchall())

    async def exq_count(self, query: SupportsGetsql) -> int:
        """Execute shorthand."""
        return len(await self.exq(cast('Query[None]', query)))

    async def __aenter__(self) -> Self:
        """Transaction setup."""
        await self.cursor.execute(f'SAVEPOINT x{id(self)}')
        return self.__class__(self.cursor)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        """Transaction cleanup."""
        if exc is not None:
            await self.cursor.execute(f'ROLLBACK TO SAVEPOINT x{id(self)}')
            raise exc
        await self.cursor.execute(f'RELEASE SAVEPOINT x{id(self)}')
        return False


class Manager:
    """Database manager."""

    def __init__(self, uri: str, options: ManagerOptions | None = None) -> None:
        """Initialize."""
        self.uri = uri
        if not options:
            options = cast('ManagerOptions', {})
        self.max_connections = options.get('max_connections', 4)
        self.pool = None

    async def init(self) -> None:
        """Initialize pool."""
        parsed = urlparse(self.uri)
        self.pool = await aiomysql.create_pool(
            host=parsed.hostname,
            port=parsed.port,
            user=parsed.username,
            password=parsed.password,
            db=parsed.path[1:],
            maxsize=self.max_connections,
            sql_mode='ANSI',
        )

    async def close(self) -> None:
        """Close all database connections."""
        assert self.pool
        self.pool.close()
        await self.pool.wait_closed()

    @asynccontextmanager
    async def txn(self, _: bool = False) -> AsyncIterator[Transaction]:
        """Open a transaction for database operations."""
        assert self.pool

        connection = await self.pool.acquire()
        cursor = await connection.cursor()
        await cursor.execute('START TRANSACTION')
        try:
            yield Transaction(cursor)
        except Exception:
            await cursor.execute('ROLLBACK')
            raise
        else:
            await cursor.execute('COMMIT')
        finally:
            await cursor.close()
            await self.pool.release(connection)



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Mysql manager tests."""

from __future__ import annotations

from contextlib import suppress
from os import getenv
from typing import TYPE_CHECKING, cast

import pytest

from sqls.transactions import IntegrityError, OperationalError, get_manager

from .common import Query

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

    from sqls.interfaces import Manager
    from sqls.transactions.mysql import Manager as MysqlManager


@pytest.fixture()
async def manager(recwarn: pytest.WarningsRecorder) -> AsyncGenerator[Manager, None]:
    """Test MySQL database manager."""
    host = getenv('SQLS_MYSQL_HOST', 'localhost')
    res = get_manager(f'mysql://sqlsuser:sqlspass@{host}:3306/sqls')
    await res.init()
    async with res.txn() as txn:
        await txn.execute('DROP TABLE IF EXISTS "foo" CASCADE')
    assert len(recwarn) <= 1
    yield res
    await res.close()


async def test_options() -> None:
    """Test options are set on manager."""
    host = getenv('SQLS_MYSQL_HOST', 'localhost')
    res = cast(
        'MysqlManager',
        get_manager(
            f'mysql://sqlsuser:sqlspass@{host}:3306/sqls',
            {'max_connections': 2},
        ),
    )
    assert res.max_connections == 2


async def test_manager(manager: MysqlManager) -> None:
    """Test Manager."""
    async with manager.txn() as txn, manager.txn() as inner_txn:
        assert inner_txn != txn

    async with manager.txn() as txn:
        await txn.execute('SELECT 1')
        assert await txn.exq_count(Query('SELECT 1')) == 1
        assert await txn.exq(Query('SELECT 1')) == [(1,)]

        assert await txn.exq(Query('SELECT ?', 1)) == [(1,)]
        assert await txn.exq(Query('SELECT ?', 1.0)) == [(1.0,)]
        assert await txn.exq(Query('SELECT ?', 'foo')) == [('foo',)]

        await txn.exq(Query('CREATE TABLE foo (bar INTEGER PRIMARY KEY)'))

    with pytest.raises(IntegrityError, match='Duplicate entry'):
        async with manager.txn() as txn:
            await txn.exq(Query('INSERT INTO foo (bar) VALUES(1),(1)'))

    with pytest.raises(OperationalError, match='error in your SQL syntax'):
        async with manager.txn() as txn:
            await txn.exq(Query('BAD SYNTAX'))

    async with manager.txn() as txn:
        with suppress(ValueError):
            async with txn as inner_txn:
                await inner_txn.exq(Query('INSERT INTO foo (bar) VALUES(1)'))
                raise ValueError
        async with txn as inner_txn:
            assert await inner_txn.exq(Query('INSERT INTO foo (bar) VALUES(1)')) == []



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL DELETE."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .base import QBase, Term

if TYPE_CHECKING:
    from .expr import Expr
    from .table import Table


class Delete(QBase[tuple[()]]):
    """DELETE mixin."""

    def delete(self, table: Table) -> PostDelete:
        """Start DELETE statement.

        Args:
            table: Table to delete.

        Returns:
            Query builder.

        """
        sql, _ = table.__getsql__()
        return self._f(PostDelete, f'DELETE FROM {sql}')  # noqa: S608


class Where(QBase[tuple[()]]):
    """WHERE mixin."""

    def where(self, expr: Expr) -> PostWhere:
        """Add WHERE clause.

        Args:
            expr: Expression.

        Returns:
            Query builder.

        """
        sql, args = expr.__getsql__()
        return self._f(PostWhere, f' WHERE {sql}', args)


class PostDelete(
    Where,
    Term[tuple[()]],
):
    """Post delete builder."""


class PostWhere(
    Term[tuple[()]],
):
    """Post where builder."""



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""CREATE TABLE tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import ExprLiteral, Query, Table
from sqls.transactions.mysql import Transaction as Mysql
from sqls.transactions.postgresql import Transaction as Postgres
from sqls.transactions.sqlite import Transaction as Sqlite

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


async def test_single_column(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('id', 'INTEGER')
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("id" INTEGER)'
    assert await txn.exq(query) == []


async def test_primary_key(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('id', 'INTEGER').primary_key()
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("id" INTEGER PRIMARY KEY)'
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        'mysql',
    ],
    indirect=True,
)
async def test_primary_key_autoincrement(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('id', 'INTEGER').primary_key(autoincrement=True)
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("id" INTEGER PRIMARY KEY AUTOINCREMENT)'
    assert await txn.exq(query) == []


async def test_not_null(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('id', 'INTEGER').not_null()
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("id" INTEGER NOT NULL)'
    assert await txn.exq(query) == []


async def test_unique(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('id', 'INTEGER').unique()
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("id" INTEGER UNIQUE)'
    assert await txn.exq(query) == []


async def test_check(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .check(tbl.id > ExprLiteral(42, bind=False))
    )
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("id" INTEGER CHECK ("tbl"."id" > 42))'
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_rollback(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('id', 'INTEGER').unique().on_conflict_rollback()
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("id" INTEGER UNIQUE ON CONFLICT ROLLBACK)'
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_abort(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('id', 'INTEGER').unique().on_conflict_abort()
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("id" INTEGER UNIQUE ON CONFLICT ABORT)'
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_fail(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('id', 'INTEGER').unique().on_conflict_fail()
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("id" INTEGER UNIQUE ON CONFLICT FAIL)'
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_ignore(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('id', 'INTEGER').unique().on_conflict_ignore()
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("id" INTEGER UNIQUE ON CONFLICT IGNORE)'
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_replace(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('id', 'INTEGER').unique().on_conflict_replace()
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("id" INTEGER UNIQUE ON CONFLICT REPLACE)'
    assert await txn.exq(query) == []


async def test_default(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('age', 'INTEGER').default(42)
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("age" INTEGER DEFAULT 42)'
    assert await txn.exq(query) == []


async def test_default_text(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('name', 'TEXT').default('foo')
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("name" TEXT DEFAULT \'foo\')'
    assert await txn.exq(query) == []


async def test_default_null(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('age', 'INTEGER').default(None)
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("age" INTEGER DEFAULT NULL)'
    assert await txn.exq(query) == []


async def test_default_true(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('age', 'INTEGER').default(True)
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("age" INTEGER DEFAULT TRUE)'
    assert await txn.exq(query) == []


async def test_default_false(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('age', 'INTEGER').default(False)
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("age" INTEGER DEFAULT FALSE)'
    assert await txn.exq(query) == []


async def test_collate(txn: Transaction) -> None:
    """Test create table statement."""
    collation = {
        Sqlite: 'NOCASE',
        Postgres: '"en_US"',
        Mysql: 'utf8_general_ci',
    }[type(txn)]
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('id', 'TEXT').collate(collation)
    assert query.__getsql__()[0] == f'CREATE TABLE "tbl" ("id" TEXT COLLATE {collation})'
    assert await txn.exq(query) == []


async def test_references(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .primary_key()
        .column('lid', 'INTEGER')
        .references(tbl, tbl.id)
    )
    assert query.__getsql__()[0] == (
        'CREATE TABLE "tbl" '
        '("id" INTEGER PRIMARY KEY,'
        ' "lid" INTEGER REFERENCES "tbl" ("id")'
        ')'
    )
    assert await txn.exq(query) == []


async def test_action_set_null_set_default(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .primary_key()
        .column('lid', 'INTEGER')
        .references(tbl, tbl.id)
        .on_delete()
        .set_null()
        .on_update()
        .set_default()
    )
    assert query.__getsql__()[0] == (
        'CREATE TABLE "tbl" '
        '("id" INTEGER PRIMARY KEY, "lid" INTEGER REFERENCES "tbl" ("id")'
        ' ON DELETE SET NULL'
        ' ON UPDATE SET DEFAULT'
        ')'
    )
    assert await txn.exq(query) == []


async def test_action_cascade_restrict(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .primary_key()
        .column('lid', 'INTEGER')
        .references(tbl, tbl.id)
        .on_delete()
        .cascade()
        .on_update()
        .restrict()
    )
    assert query.__getsql__()[0] == (
        'CREATE TABLE "tbl" '
        '("id" INTEGER PRIMARY KEY, "lid" INTEGER REFERENCES "tbl" ("id")'
        ' ON DELETE CASCADE'
        ' ON UPDATE RESTRICT'
        ')'
    )
    assert await txn.exq(query) == []


async def test_no_action(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .primary_key()
        .column('lid', 'INTEGER')
        .references(tbl, tbl.id)
        .on_delete()
        .no_action()
    )
    assert query.__getsql__()[0] == (
        'CREATE TABLE "tbl" '
        '("id" INTEGER PRIMARY KEY, "lid" INTEGER REFERENCES "tbl" ("id")'
        ' ON DELETE NO ACTION'
        ')'
    )
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_deferrable(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .primary_key()
        .column('lid', 'INTEGER')
        .references(tbl, tbl.id)
        .deferrable()
    )
    assert query.__getsql__()[0] == (
        'CREATE TABLE "tbl" '
        '("id" INTEGER PRIMARY KEY, "lid" INTEGER REFERENCES "tbl" ("id")'
        ' DEFERRABLE'
        ')'
    )
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_not_deferrable(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .primary_key()
        .column('lid', 'INTEGER')
        .references(tbl, tbl.id)
        .deferrable(is_deferrable=False)
    )
    assert query.__getsql__()[0] == (
        'CREATE TABLE "tbl" '
        '("id" INTEGER PRIMARY KEY, "lid" INTEGER REFERENCES "tbl" ("id")'
        ' NOT DEFERRABLE'
        ')'
    )
    assert await txn.exq(query) == []


async def test_generated(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    anon = Table('')
    query = (
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .column('lid', 'INTEGER')
        .generated(anon.id + ExprLiteral(1, bind=False))
    )
    assert query.__getsql__()[0] == (
        'CREATE TABLE "tbl" '
        '("id" INTEGER, "lid" INTEGER'
        ' GENERATED ALWAYS AS ("id" + 1) VIRTUAL'
        ')'
    )
    assert await txn.exq(query) == []


async def test_unique_columns(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .column('lid', 'INTEGER')
        .unique_columns('id', 'lid')
    )
    assert (
        query.__getsql__()[0]
        == 'CREATE TABLE "tbl" ("id" INTEGER, "lid" INTEGER, UNIQUE (id, lid))'
    )
    assert await txn.exq(query) == []


async def test_check_columns(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .check_columns(tbl.id > ExprLiteral(42, bind=False))
    )
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("id" INTEGER, CHECK ("tbl"."id" > 42))'
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        'mysql',
    ],
    indirect=True,
)
async def test_foreign_key(txn: Transaction) -> None:
    """Test create table statement."""
    dst = Table('dst')
    await txn.exq(
        Query()
        .create_table(dst)
        .column('id', 'INTEGER')
        .column('lid', 'INTEGER')
        .unique()
        .column('mid', 'INTEGER')
        .unique()
        .unique_columns('lid', 'mid'),
    )

    tbl = Table('tbl')
    query = (
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .column('lid', 'INTEGER')
        .column('mid', 'INTEGER')
        .foreign_key('lid', 'mid')
        .references(dst, tbl.lid, tbl.mid)
        .on_delete()
        .cascade()
    )
    assert query.__getsql__()[0] == (
        'CREATE TABLE "tbl" '
        '("id" INTEGER, "lid" INTEGER, "mid" INTEGER,'
        ' FOREIGN KEY (lid, mid) REFERENCES "dst" ("lid", "mid") ON DELETE CASCADE'
        ')'
    )
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='unimplemented')),
    ],
    indirect=True,
)
async def test_foreign_key_deferrable(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .column('lid', 'INTEGER')
        .column('mid', 'INTEGER')
        .foreign_key('lid', 'mid')
        .references(tbl, tbl.mid, tbl.lid)
        .deferrable()
        .unique_columns('lid', 'mid')
    )
    assert query.__getsql__()[0] == (
        'CREATE TABLE "tbl" '
        '("id" INTEGER, "lid" INTEGER, "mid" INTEGER,'
        ' FOREIGN KEY (lid, mid) REFERENCES "tbl" ("mid", "lid") DEFERRABLE,'
        ' UNIQUE (lid, mid)'
        ')'
    )
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_without_rowid(txn: Transaction) -> None:
    """Test create table statement."""
    tbl = Table('tbl')
    query = Query().create_table(tbl).column('id', 'INTEGER').primary_key().without_rowid()
    assert query.__getsql__()[0] == 'CREATE TABLE "tbl" ("id" INTEGER PRIMARY KEY) WITHOUT ROWID'
    assert await txn.exq(query) == []



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""VACUUM tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import ExprLiteral, Query, Table

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_simple(txn: Transaction) -> None:
    """Test vacuum statement."""
    await txn.execute('ROLLBACK')

    query = Query().vacuum()
    assert query.__getsql__() == ('VACUUM', ())
    await txn.exq(query)

    await txn.execute('BEGIN TRANSACTION')


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_schema_or_table(txn: Transaction) -> None:
    """Test vacuum statement."""
    tbl = Table('main')
    await txn.exq(Query().create_table(tbl).column('id', 'INTEGER'))
    await txn.execute('COMMIT')

    query = Query().vacuum(schema='main')
    assert query.__getsql__() == ('VACUUM "main"', ())
    await txn.exq(query)

    await txn.execute('BEGIN TRANSACTION')


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_into(txn: Transaction) -> None:
    """Test vacuum statement."""
    tbl = Table('main')
    await txn.exq(Query().create_table(tbl).column('id', 'INTEGER'))
    await txn.execute('COMMIT')

    query = Query().vacuum(schema='main', into='file:bar.db?mode=memory')
    assert query.__getsql__() == ('VACUUM "main" INTO ?', ('file:bar.db?mode=memory',))
    await txn.exq(query)

    query = Query().vacuum(into=ExprLiteral('file:bar').strconcat('.db?mode=memory'))
    assert query.__getsql__() == ('VACUUM INTO ? || ?', ('file:bar', '.db?mode=memory'))
    await txn.exq(query)

    await txn.execute('BEGIN TRANSACTION')



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""DELETE FROM tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import Query, Table

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.fixture(autouse=True)
async def _autotable(txn: Transaction) -> None:
    tbl = Table('tbl')
    await txn.exq(
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .primary_key()
        .column('name', 'TEXT')
        .unique()
        .column('age', 'INTEGER'),
    )


async def test_delete(txn: Transaction) -> None:
    """Test delete statement."""
    tbl = Table('tbl')
    delete = Query().delete(tbl)
    assert delete.__getsql__() == ('DELETE FROM "tbl"', ())
    await txn.exq(delete)

    delete_where = Query().delete(tbl).where(tbl.name == 'foo')
    assert delete_where.__getsql__() == ('DELETE FROM "tbl" WHERE "tbl"."name" = ?', ('foo',))
    await txn.exq(delete_where)



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL ATTACH."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .base import QBase, Term

if TYPE_CHECKING:
    from .expr import Expr


class Attach(QBase[tuple[()]]):
    """ATTACH DATABASE mixin."""

    def attach(self, expr: Expr, schema: str) -> Term[tuple[()]]:
        """Create ATTACH DATABASE statement.

        Args:
            expr: Database to attach.
            schema: Schema name to attach to.

        Returns:
            Query.

        """
        esql, args = expr.__getsql__()
        schema = schema.replace('"', '""')
        sql = f'ATTACH DATABASE {esql} AS "{schema}"'
        return self._f(Term, sql, args)



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL tables and joins."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .expr import Expr, ExprColumn

if TYPE_CHECKING:
    from typing import ClassVar, TypeVar

    from sqls.interfaces import BindArg, SqlValue
    from sqls.queries.base import QBase

    T = TypeVar('T', bound='JoinBase')


class Tabular:
    """Table base class."""

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        raise NotImplementedError


class Aliasable(Tabular):
    """Aliasable entity."""

    def as_(self, alias: str | Table) -> TableAlias:
        """Alias table.

        Args:
            alias: New table name.

        Returns:
            Aliased table.

        """
        return TableAlias(self, alias)

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        raise NotImplementedError  # pragma: no cover


class Table(Aliasable):
    """Database table."""

    def __init__(self, name: str, schema: str | None = None) -> None:
        """Initialize table.

        Args:
            name: Table name.
            schema: Schema name.

        """
        super().__init__()
        self._table = name
        self._schema = schema
        self.__star__ = ExprColumn(schema, name, '*')

    def __truediv__(self, key: str) -> ExprColumn:
        """Get column with expression '/' operator.

        Args:
            key: Column name.

        Returns:
            Column expression.

        """
        return ExprColumn(self._schema, self._table, key)

    def __getattr__(self, key: str) -> ExprColumn:
        """Get column expression from attribute access.

        Args:
            key: Column name.

        Returns:
            Column expression.

        """
        return ExprColumn(self._schema, self._table, key)

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        table = self._table.replace('"', '""')
        if self._schema:
            schema = self._schema.replace('"', '""')
            return f'"{schema}"."{table}"', ()
        return f'"{table}"', ()


class TableQuery(Aliasable):
    """Subquery used as table."""

    def __init__(self, query: QBase[tuple[SqlValue | object, ...]]) -> None:
        """Initialize table from subquery.

        Args:
            query: Subquery.

        """
        super().__init__()
        self.query = query

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        sql, args = self.query.__getsql__()
        return f'({sql})', args


class TableAlias(Tabular):
    """Tabular with applied alias."""

    def __init__(self, table: Tabular, alias: str | Table) -> None:
        """Initialize.

        Args:
            table: Table or TableQuery.
            alias: New table name.

        """
        super().__init__()
        self._tabular = table
        self._alias: str = alias if isinstance(alias, str) else alias._table  # noqa: SLF001
        self.__star__ = ExprColumn('', self._alias, '*')

    def __truediv__(self, key: str) -> ExprColumn:
        """Get column with expression '/' operator.

        Args:
            key: Column name.

        Returns:
            Column expression.

        """
        return ExprColumn(getattr(self._tabular, '_schema', ''), self._alias, key)

    def __getattr__(self, key: str) -> ExprColumn:
        """Get column expression from attribute access.

        Args:
            key: Column name.

        Returns:
            Column expression.

        """
        return ExprColumn(getattr(self._tabular, '_schema', ''), self._alias, key)

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        sql, args = self._tabular.__getsql__()
        tabular = getattr(self, '_tabular')  # noqa: B009
        if isinstance(tabular, JoinTerm):
            sql = f'({sql})'
        return f'{sql} AS "{self._alias}"', args


class JoinBase:
    """Base for JOIN clause builders."""

    _base: Tabular | JoinBase | None
    _joins: list[tuple[str, Tabular | JoinBase, Expr | None]]
    types: ClassVar[dict[str | None, str]] = {
        'cross': 'CROSS JOIN',
        'inner': 'INNER JOIN',
        'left': 'LEFT JOIN',
        None: 'JOIN',
    }

    def __init__(self, parent: JoinBase | None = None) -> None:
        """Initialize.

        Args:
            parent: Parent join object.

        """
        super().__init__()
        if parent:
            self._base = parent._base  # noqa: SLF001
            self._joins = parent._joins[:]  # noqa: SLF001
        else:
            self._base = None
            self._joins = []

    def _forward_join(self, cls: type[T], join: Tabular | JoinBase, typ: str) -> T:
        """Apply JOIN clause.

        Args:
            cls: Class of next builder in chain.
            join: Table or other Join.
            typ: Join type.

        Returns:
            Instance of next builder.

        """
        inst = cls(self)
        if self._base:
            inst._joins.append((typ, join, None))  # noqa: SLF001
        else:
            inst._base = join  # noqa: SLF001

        return inst

    def _forward_constraint(self, cls: type[T], expr: Expr) -> T:
        """Apply ON clause.

        Args:
            cls: Class of next builder in chain.
            expr: Expression to join on.

        Returns:
            Instance of next builder.

        """
        inst = cls(self)
        typ, table, _ = self._joins[-1]
        inst._joins[-1] = (typ, table, expr)  # noqa: SLF001
        return inst

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Get SQL string and args."""
        assert self._base
        sql, args = self._base.__getsql__()
        for typ, table, expr in self._joins:
            sql += f' {typ} '
            subsql, subargs = table.__getsql__()
            sql += subsql
            args += subargs
            if expr:
                sql += ' ON '
                subsql, subargs = expr.__getsql__()
                sql += subsql
                args += subargs

        return sql, args


class Join(JoinBase):
    """JOIN builder."""

    def join(self, tabular: Tabular) -> JoinTerm:
        """Add first table to join.

        Args:
            tabular: Table, Subquery, or other Join.

        Returns:
            Join builder.

        """
        return self._forward_join(JoinTerm, tabular, '')


class JoinTerm(JoinBase, Aliasable):
    """Initial join state."""

    def join(self, tabular: Tabular, typ: str | None = None) -> JoinConstrainable:
        """JOIN tabular to current state.

        Args:
            tabular: Table, Subquery, or other Join.
            typ: Join type.

        Returns:
            Join builder.

        """
        return self._forward_join(JoinConstrainable, tabular, self.types[typ])


class JoinConstrainable(JoinTerm):
    """Constrainable join state."""

    def on_(self, expr: Expr) -> JoinTerm:
        """Constrain last JOIN with ON clause.

        Args:
            expr: Expression to join on.

        Returns:
            Join builder.

        """
        return self._forward_constraint(JoinTerm, expr)



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL CREATE VIEW."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .base import QBase, Term

if TYPE_CHECKING:
    from .base import Texplicit
    from .expr import ExprColumn
    from .table import Table


class CreateView(QBase[tuple[()]]):
    """CREATE VIEW mixin."""

    def create_view(
        self,
        view: Table,
        temp: bool = False,
        if_not_exists: bool = False,
    ) -> PostCreate:
        """Start CREATE VIEW statement.

        Args:
            view: Table to create.
            temp: Is view temporary.
            if_not_exists: Ignore existing view.

        Returns:
            Query builder.

        """
        sql, _ = view.__getsql__()
        tmp = ' TEMP' if temp else ''
        ine = ' IF NOT EXISTS' if if_not_exists else ''
        sql = f'CREATE{tmp} VIEW{ine} {sql}'
        return self._f(PostCreate, sql)


class Columns(QBase[tuple[()]]):
    """Columns mixin."""

    def columns(self, *cols: ExprColumn) -> PostColumns:
        """Add columns.

        Args:
            cols: Columns.

        Returns:
            Query builder.

        """
        return self._f(PostColumns, f'({", ".join(x.col for x in cols)})')


class As(QBase[tuple[()]]):
    """AS mixin."""

    def as_(self, select: QBase[Texplicit]) -> PostAs:
        """Add AS clause.

        Args:
            select: Select statement.

        Returns:
            Query builder.

        """
        sql, args = select.__getsql__()
        return self._f(PostAs, f' AS {sql}', args)


class PostCreate(
    Columns,
    As,
):
    """Post create builder."""


class PostColumns(
    As,
):
    """Post columns builder."""


class PostAs(
    Term[tuple[()]],
):
    """Post AS builder."""



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL WITH."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .base import QBase, Term
from .delete import Delete
from .insert import Insert
from .select import Select
from .update import Update

if TYPE_CHECKING:
    from .base import Texplicit
    from .expr import ExprColumn
    from .table import Table


class With(QBase[tuple[()]]):
    """WITH mixin."""

    def with_(self, table: Table, *cols: ExprColumn, recursive: bool = False) -> As:
        """Start WITH clause.

        Args:
            table: Table.
            cols: Table columns.
            recursive: Allow recursive queries.

        Returns:
            Query builder.

        """
        sql, _ = table.__getsql__()
        rec = ' RECURSIVE' if recursive else ''
        sql = f'WITH{rec} {sql} ({", ".join(x.col for x in cols)})'
        return self._f(As, sql)


class As(QBase[tuple[()]]):
    """AS mixin for WITH."""

    def as_(self, expr: QBase[Texplicit]) -> PostAs:
        """Add AS clause.

        Args:
            expr: Common table expression.

        Returns:
            Query builder.

        """
        sql, args = expr.__getsql__()
        return self._f(PostAs, f' AS ({sql})', args)


class AddWith(QBase[tuple[()]]):
    """WITH mixin for adding additional with clauses."""

    def with_(self, table: Table, *cols: ExprColumn) -> As:
        """Start WITH clause.

        Args:
            table: Table.
            cols: Table columns.

        Returns:
            Query builder.

        """
        sql, _ = table.__getsql__()
        return self._f(As, f', {sql} ({", ".join(x.col for x in cols)})')


class PostAs(
    AddWith,
    Delete,
    Insert,
    Select,
    Update,
    Term[tuple[()]],
):
    """Post AS Builder."""



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""CREATE INDEX tests."""

from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING

import pytest

from sqls.queries import ExprLiteral, Query, Table
from sqls.transactions import IntegrityError

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.fixture(autouse=True)
async def _autotable(txn: Transaction) -> None:
    tbl = Table('tbl')
    await txn.exq(
        Query().create_table(tbl).column('id', 'INTEGER').primary_key().column('name', 'TEXT'),
    )


async def test_simple(txn: Transaction) -> None:
    """Test create index statement."""
    tbl = Table('tbl')

    assert (
        await txn.exq(
            Query().create_index(Table('idx'), tbl, tbl.name),
        )
        == []
    )


async def test_if_not_exists(txn: Transaction) -> None:
    """Test create index statement."""
    tbl = Table('tbl')

    assert (
        await txn.exq(
            Query().create_index(Table('idx'), tbl, tbl.name, if_not_exists=True),
        )
        == []
    )


async def test_unique(txn: Transaction) -> None:
    """Test create index statement."""
    tbl = Table('tbl')

    assert (
        await txn.exq(
            Query().create_index(Table('idx'), tbl, tbl.name, unique=True),
        )
        == []
    )

    query = Query().insert(tbl).columns(tbl.name).values(('foo',))
    assert await txn.exq(query) == []

    with suppress(IntegrityError):
        async with txn as subtxn:  # type: ignore[attr-defined]
            with pytest.raises(IntegrityError) as err:
                await subtxn.exq(query)
            assert 'unique constraint' in str(err).lower() or 'duplicate entry' in str(err).lower()
            raise IntegrityError


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip('unsupported')),
    ],
    indirect=True,
)
async def test_partial(txn: Transaction) -> None:
    """Test create index statement."""
    tbl = Table('tbl')

    assert (
        await txn.exq(
            Query()
            .create_index(Table('pidx'), tbl, tbl.name)
            .where(tbl.name == ExprLiteral('42', bind=False)),
        )
        == []
    )



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SELECT tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import ExprFunction, ExprLiteral, Query, QueryError, Table

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.fixture(autouse=True)
async def _autotable(txn: Transaction) -> None:
    tbl = Table('tbl')
    await txn.exq(
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .primary_key(autoincrement=True)
        .unique()
        .column('name', 'TEXT')
        .unique()
        .column('age', 'INTEGER'),
    )

    await txn.exq(
        Query()
        .insert(tbl)
        .columns(tbl.name, tbl.age)
        .values(
            ('John', 42),
            ('Paul', 43),
            ('Ringo', 44),
            ('George', 45),
        ),
    )


async def test_values(txn: Transaction) -> None:
    """Test select statement."""
    with pytest.raises(QueryError, match='one row'):
        Query().values().__getsql__()

    with pytest.raises(QueryError, match='one column'):
        Query().values(()).__getsql__()

    query = Query().values((1, 2), (3, 4))
    assert query.__getsql__() == ('VALUES(?, ?),(?, ?)', (1, 2, 3, 4))
    assert await txn.exq(query) == [(1, 2), (3, 4)]


async def test_from_table(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    query = Query().select(tbl.id.typed(int)).from_(tbl)
    assert query.__getsql__() == ('SELECT "tbl"."id" FROM "tbl"', ())
    assert await txn.exq(query) == [(1,), (2,), (3,), (4,)]


async def test_from_subquery(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .select(Table('').__star__.typed(object))
        .from_(
            Query().select(tbl.id.typed(int)).from_(tbl).table().as_('sub'),
        )
    )
    assert query.__getsql__() == ('SELECT * FROM (SELECT "tbl"."id" FROM "tbl") AS "sub"', ())
    assert await txn.exq(query) == [(1,), (2,), (3,), (4,)]


async def test_where(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    query = Query().select(tbl.id.typed(int)).from_(tbl).where(None)
    assert query.__getsql__() == ('SELECT "tbl"."id" FROM "tbl"', ())
    assert await txn.exq(query) == [(1,), (2,), (3,), (4,)]

    query = Query().select(tbl.id.typed(int)).from_(tbl).where(tbl.name == 'Ringo')
    assert query.__getsql__() == ('SELECT "tbl"."id" FROM "tbl" WHERE "tbl"."name" = ?', ('Ringo',))
    assert await txn.exq(query) == [(3,)]


async def test_function(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    query = (
        Query().select(ExprFunction('COUNT', tbl.id).typed(int)).from_(tbl).group_by(tbl.age > 44)
    )
    assert query.__getsql__() == (
        'SELECT COUNT("tbl"."id") FROM "tbl" GROUP BY "tbl"."age" > ?',
        (44,),
    )
    assert await txn.exq(query) == [(3,), (1,)]


async def test_groupby(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .select(
            ExprFunction('LENGTH', tbl.name).typed(int),
            ExprFunction('COUNT', tbl.id).typed(int),
        )
        .from_(tbl)
        .group_by(ExprFunction('LENGTH', tbl.name))
        .having(ExprFunction('COUNT', tbl.id) > 1)
    )
    assert query.__getsql__() == (
        'SELECT LENGTH("tbl"."name"), COUNT("tbl"."id") FROM "tbl" GROUP BY '
        'LENGTH("tbl"."name") HAVING (COUNT("tbl"."id")) > ?',
        (1,),
    )
    assert await txn.exq(query) == [(4, 2)]


async def test_union(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .select(tbl.name.typed(str))
        .from_(tbl)
        .where(tbl.age >= 43)
        .union()
        .select(tbl.name.typed(str))
        .from_(tbl)
        .where(tbl.age < 45)
    )
    assert query.__getsql__() == (
        'SELECT "tbl"."name" FROM "tbl" WHERE "tbl"."age" >= ? UNION '
        'SELECT "tbl"."name" FROM "tbl" WHERE "tbl"."age" < ?',
        (43, 45),
    )
    assert sorted(await txn.exq(query)) == [
        ('George',),
        ('John',),
        ('Paul',),
        ('Ringo',),
    ]


async def test_union_query(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .select(tbl.name.typed(str))
        .from_(tbl)
        .where(tbl.age >= 43)
        .union_query(
            Query().select(tbl.name.typed(str)).from_(tbl).where(tbl.age < 45),
        )
    )
    assert query.__getsql__() == (
        'SELECT "tbl"."name" FROM "tbl" WHERE "tbl"."age" >= ? UNION '
        'SELECT "tbl"."name" FROM "tbl" WHERE "tbl"."age" < ?',
        (43, 45),
    )
    assert sorted(await txn.exq(query)) == [
        ('George',),
        ('John',),
        ('Paul',),
        ('Ringo',),
    ]


async def test_union_all(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .select(tbl.name.typed(str))
        .from_(tbl)
        .where(tbl.age >= 43)
        .union_all()
        .select(tbl.name.typed(str))
        .from_(tbl)
        .where(tbl.age < 45)
    )
    assert query.__getsql__() == (
        'SELECT "tbl"."name" FROM "tbl" WHERE "tbl"."age" >= ? UNION ALL '
        'SELECT "tbl"."name" FROM "tbl" WHERE "tbl"."age" < ?',
        (43, 45),
    )
    assert await txn.exq(query) == [
        ('Paul',),
        ('Ringo',),
        ('George',),
        ('John',),
        ('Paul',),
        ('Ringo',),
    ]


async def test_intersect(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .select(tbl.name.typed(str))
        .from_(tbl)
        .where(tbl.age >= 43)
        .intersect()
        .select(tbl.name.typed(str))
        .from_(tbl)
        .where(tbl.age < 45)
    )
    assert query.__getsql__() == (
        'SELECT "tbl"."name" FROM "tbl" WHERE "tbl"."age" >= ? INTERSECT '
        'SELECT "tbl"."name" FROM "tbl" WHERE "tbl"."age" < ?',
        (43, 45),
    )
    assert sorted(await txn.exq(query)) == [
        ('Paul',),
        ('Ringo',),
    ]


async def test_except(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .select(tbl.name.typed(str))
        .from_(tbl)
        .where(tbl.age >= 43)
        .except_()
        .select(tbl.name.typed(str))
        .from_(tbl)
        .where(tbl.age < 45)
    )
    assert query.__getsql__() == (
        'SELECT "tbl"."name" FROM "tbl" WHERE "tbl"."age" >= ? EXCEPT '
        'SELECT "tbl"."name" FROM "tbl" WHERE "tbl"."age" < ?',
        (43, 45),
    )
    assert sorted(await txn.exq(query)) == [
        ('George',),
    ]


async def test_orderby(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    query = Query().select(tbl.name.typed(str)).from_(tbl).order_by(tbl.name)
    assert query.__getsql__() == (
        'SELECT "tbl"."name" FROM "tbl" ORDER BY "tbl"."name"',
        (),
    )
    assert await txn.exq(query) == [
        ('George',),
        ('John',),
        ('Paul',),
        ('Ringo',),
    ]

    query = Query().select(tbl.name.typed(str)).from_(tbl).order_by(tbl.name, desc=True)
    assert query.__getsql__() == (
        'SELECT "tbl"."name" FROM "tbl" ORDER BY "tbl"."name" DESC',
        (),
    )
    assert await txn.exq(query) == [
        ('Ringo',),
        ('Paul',),
        ('John',),
        ('George',),
    ]


async def test_limit(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    query = Query().select(tbl.name.typed(str)).from_(tbl).order_by(tbl.name).limit(1).offset(1)
    assert query.__getsql__() == (
        'SELECT "tbl"."name" FROM "tbl" ORDER BY "tbl"."name" LIMIT ? OFFSET ?',
        (1, 1),
    )
    assert await txn.exq(query) == [
        ('John',),
    ]

    query = Query().select(tbl.name.typed(str)).from_(tbl).order_by(tbl.name).limit(2).offset(2)
    assert query.__getsql__() == (
        'SELECT "tbl"."name" FROM "tbl" ORDER BY "tbl"."name" LIMIT ? OFFSET ?',
        (2, 2),
    )
    assert await txn.exq(query) == [
        ('Paul',),
        ('Ringo',),
    ]


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_limit_expr(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .select(tbl.name.typed(str))
        .from_(tbl)
        .order_by(tbl.name)
        .limit(ExprLiteral(1) + 1)
        .offset(ExprLiteral(1) + 1)
    )
    assert query.__getsql__() == (
        'SELECT "tbl"."name" FROM "tbl" ORDER BY "tbl"."name" LIMIT ? + ? OFFSET ? + ?',
        (1, 1, 1, 1),
    )
    assert await txn.exq(query) == [
        ('Paul',),
        ('Ringo',),
    ]


async def test_with(txn: Transaction) -> None:
    """Test select statement."""
    tbl = Table('tbl')
    dst = Table('dst')
    query = (
        Query()
        .with_(tbl, tbl.id, tbl.id2)
        .as_(
            Query().values((1, 2)),
        )
        .with_(dst, dst.id, dst.id2)
        .as_(
            Query().values((3, 4)),
        )
        .select(tbl.__star__.typed(object))
        .from_(tbl)
        .union()
        .select(dst.__star__.typed(object))
        .from_(dst)
    )
    assert query.__getsql__() == (
        'WITH "tbl" (id, id2) AS (VALUES(?, ?)), "dst" (id, id2) AS (VALUES(?, ?)) '
        'SELECT "tbl".* FROM "tbl" UNION SELECT "dst".* FROM "dst"',
        (1, 2, 3, 4),
    )
    assert await txn.exq(query) == [(1, 2), (3, 4)]  # type: ignore[comparison-overlap]



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Table tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import Join, Query, Table
from sqls.queries.table import Tabular

if TYPE_CHECKING:
    from typing import TypeVar

    from sqls.interfaces import (
        Query as QueryProtocol,
        Transaction,
    )

    T = TypeVar('T')


async def test_table(txn: Transaction) -> None:
    """Test tables."""
    with pytest.raises(NotImplementedError):
        Tabular().__getsql__()

    tbl = Table('tbl', schema='foo')
    assert tbl.__getsql__()[0] == '"foo"."tbl"'
    assert tbl.as_('foo').__getsql__()[0] == '"foo"."tbl" AS "foo"'

    tbl = Table('tbl')
    assert tbl.__getsql__()[0] == '"tbl"'
    assert tbl.as_('foo').__getsql__()[0] == '"tbl" AS "foo"'

    assert tbl.bar.__getsql__() == ('"tbl"."bar"', ())
    assert (tbl / 'bar').__getsql__() == ('"tbl"."bar"', ())
    assert tbl.as_('foo').bar.__getsql__() == ('"foo"."bar"', ())
    assert (tbl.as_('foo') / 'bar').__getsql__() == ('"foo"."bar"', ())

    sub = Table('sub')

    assert Join().join(tbl).join(sub).__getsql__() == ('"tbl" JOIN "sub"', ())
    assert Join().join(tbl).join(sub).as_('foo').__getsql__() == ('("tbl" JOIN "sub") AS "foo"', ())

    async def exc(expr: QueryProtocol[T]) -> list[T]:
        return await txn.exq(expr)

    query = (
        Query()
        .with_(tbl, tbl.id, tbl.name)
        .as_(
            Query().values((1, 'foo'), (2, 'bar')),
        )
        .with_(sub, sub.id, sub.name)
        .as_(
            Query().values((1, 'baz'), (3, 'zar')),
        )
        .select(tbl.id.typed(int), tbl.name.typed(str), sub.id.typed(int), sub.name.typed(str))
        .from_(Join().join(tbl).join(sub, typ='cross'))
    )
    assert sorted(await exc(query)) == [
        (1, 'foo', 1, 'baz'),
        (1, 'foo', 3, 'zar'),
        (2, 'bar', 1, 'baz'),
        (2, 'bar', 3, 'zar'),
    ]

    alias = Table('alias')

    query = (
        Query()
        .with_(tbl, tbl.id, tbl.name)
        .as_(
            Query().values((1, 'foo'), (2, 'bar')),
        )
        .with_(sub, sub.id, sub.name)
        .as_(
            Query().values((1, 'baz'), (3, 'zar')),
        )
        .select(alias.id.typed(int), alias.name.typed(str), sub.id.typed(int), sub.name.typed(str))
        .from_(
            Join()
            .join(tbl.as_('alias'))
            .join(tbl, typ='cross')
            .join(sub)
            .on_(alias.id == sub.id)
            .join(sub.as_('other'), typ='cross'),
        )
    )
    assert await exc(query) == [
        (1, 'foo', 1, 'baz'),
        (1, 'foo', 1, 'baz'),
        (1, 'foo', 1, 'baz'),
        (1, 'foo', 1, 'baz'),
    ]



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL ALTER."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .base import QBase, Term
from .create_table import (
    ColumnConstraintGeneric,
    DeferrableGeneric,
    OnConflictGeneric,
    OnRefchangeActionGeneric,
    OnRefchangeGeneric,
)

if TYPE_CHECKING:
    from .expr import ExprColumn
    from .table import Table


class Alter(QBase[tuple[()]]):
    """ALTER TABLE mixin."""

    def alter_table(self, table: Table) -> PostAlter:
        """Start ALTER TABLE statement."""
        return self._f(PostAlter, f'ALTER TABLE {table.__getsql__()[0]}')


class Rename(QBase[tuple[()]]):
    """RENAME mixin."""

    def rename_to(self, table: Table) -> PostRename:
        """Add RENAME TO clause.

        Args:
            table: Target table.

        Returns:
            Query builder.

        """
        return self._f(PostRename, f' RENAME TO {table.__getsql__()[0]}')

    def rename_column(self, old: ExprColumn, new: ExprColumn) -> PostRename:
        """Add RENAME COLUMN clause.

        Args:
            old: Old column.
            new: New column.

        Returns:
            Query builder.

        """
        oldcol = old.col.replace('"', '""')
        newcol = new.col.replace('"', '""')
        return self._f(PostRename, f' RENAME COLUMN "{oldcol}" TO "{newcol}"')


class Drop(QBase[tuple[()]]):
    """DROP mixin."""

    def drop_column(self, col: ExprColumn) -> PostDrop:
        """Add DROP COLUMN clause.

        Args:
            col: Column.

        Returns:
            Query builder.

        """
        newcol = col.col.replace('"', '""')
        return self._f(PostDrop, f' DROP COLUMN "{newcol}"')


class Add(QBase[tuple[()]]):
    """ADD COLUMN mixin."""

    def add_column(self, column: ExprColumn, typ: str | None) -> PostAdd:
        """Start ADD COLUMN clause.

        Args:
            column: Column.
            typ: Column type.

        Returns:
            Query builder.

        """
        col = column.col.replace('"', '""')
        typ = f' {typ}' if typ else ''
        return self._f(PostAdd, f' ADD COLUMN {col}{typ}')


class ColumnConstraint(
    ColumnConstraintGeneric[
        'PostColumnConstraint',
        'PostAdd',
        'PostColumnConstraintReferences',
    ],
):
    """Column constraint."""


class OnRefchange(OnRefchangeGeneric['OnRefchangeAction']):
    """On reference change mixin."""


class OnRefchangeAction(OnRefchangeActionGeneric['PostColumnConstraintReferences']):
    """On reference change action mixin."""


class Deferrable(DeferrableGeneric['PostAdd']):
    """Deferrable mixin."""


class ConflictClause(OnConflictGeneric['PostAdd']):
    """Conflict clause mixin."""


class PostAlter(
    Rename,
    Drop,
    Add,
):
    """Post ALTER builder."""


class PostRename(
    Term[tuple[()]],
):
    """Post RENAME builder."""


class PostDrop(
    Term[tuple[()]],
):
    """Post DROP builder."""


class PostAdd(
    ColumnConstraint,
    Term[tuple[()]],
):
    """Post ADD builder."""


class PostColumnConstraint(
    PostAdd,
    ConflictClause,
):
    """Post ColumnConstraint builder."""


class PostColumnConstraintReferences(
    PostAdd,
    OnRefchange,
    Deferrable,
):
    """Post ColumnConstraintReferences builder."""


class PostConflictClause(
    PostAdd,
):
    """Post ConflictClause builder."""
